
(* SML.Net sample program for neural network XOR problem using fannKernel *)

structure fannSML
  : sig val main: string option array option -> unit
    end = 
struct
    fun for(start, stop, f) =
        let
            fun loop i =
            if i > stop
                then ()
            else (f i; loop (i + 1))
        in
            loop start
        end

    fun ri2s(r: real): string =
        String.translate
            (fn #"~" => "-" | c => str c)
            (Int.toString(Real.round(r)))

    fun r2s(r: real): string =
        if Real.== (r, 0.0) then "0.000000"
        else String.translate
            (fn #"~" => "-" | c => str c)
            (Real.fmt (StringCvt.FIX (SOME 6)) r)

    fun xorCreate(fann:fannDotNet.fannProxy.fannKernel, net:int) =
        let
            open fannDotNet.fannProxy
            val layers = Array.fromList [2, 3, 1]
        in
            fann.#Create(net, NetworkType.Layer, layers, 1.0 : Real32.real);
            fann.#SetLearningRate(net, 0.7 : Real32.real);
            fann.#SetActivationSteepnessHidden(net, 1.0);
            fann.#SetActivationSteepnessOutput(net, 1.0);
            fann.#SetActivationFunctionHidden(net, ActivationFunction.SigmoidSymmetricStepwise);
            fann.#SetActivationFunctionOutput(net, ActivationFunction.SigmoidSymmetricStepwise);
            fann.#PrintParameters(net) (* Displays the network parameters in fannKernel *)
        end

    fun xorTrain(fann:fannDotNet.fannProxy.fannKernel, net:int) =
        let
            val epochs = Word.fromInt 300
            val reportInterval = Word.fromInt 30
            val stopAtMSE = 0.001 : Real32.real
            val mse = fann.#TrainOnData(net, epochs, reportInterval, stopAtMSE)
        in
            print "Output the mean square error information\n";
            print "\tEpoch\t\tMSE\n";

            for (0, Array.length(valOf mse) - 1, fn i =>
                let
                    val e = valOf(Array.sub(valOf mse, i))
                in
                    print("\t" ^ ri2s(Array.sub(e, 0)));
                    print("\t\t" ^ r2s(Array.sub(e, 1)) ^ "\n")
                end)
        end

    fun xorExecute(fann:fannDotNet.fannProxy.fannKernel, net:int) =
        let
            val data = Array.fromList [
                    SOME(Array.fromList [~1.0, ~1.0]),
                    SOME(Array.fromList [~1.0,  1.0]),
                    SOME(Array.fromList [ 1.0, ~1.0]),
                    SOME(Array.fromList [ 1.0,  1.0]) ]
            val result = fann.#RunData(net, data)
        in
            print "Output the test data and results\n";
            print "\tData\t\tResult\n";
            
            for (0, Array.length(data) - 1, fn i =>
                let
                    val d = valOf(Array.sub(data, i))
                    val r = valOf(Array.sub(valOf result, 0))
                in
                    print("\t");
                    Array.app (fn di => print (ri2s di ^ "\t")) d;
                    print (r2s(Array.sub(r, i)) ^ "\n")
                end)
        end

    fun xorSample() =
        let
            open fannDotNet.fannProxy
            val fann = fannKernel()
            val net = fann.#CreateContext()
        in
            print "Create the neural network and set its parameters\n";
            xorCreate(fann, net);

            print "Read training data from the xor.train sample file\n";
            fann.#ReadTrainFromFile(net, "xor.train");

            print "Initialize and train the neural network\n";
            fann.#InitWeights net;
            xorTrain(fann, net);

            print "Execute the neural network on test data\n";
            xorExecute(fann, net);

            print "Remove the neural network from fannKernel\n";
            fann.#DestroyContext net
        end

    fun main (a : string option array option) =
    (
        print "SML XOR Sample - only works if fannKernel has been started...\n";
        print "Connecting to fannKernel and creating a neural network context\n";
        xorSample()
    )
    handle _ => print("\nSML XOR Sample Exception\n")
end
