// Copyright 2005/2548 written by freegoldbar at yahoo dot com.
// All rights reserved. For private and educational use.
// Not for sale, commercial or military use.
// No warranties given. No suitability for any purpose.

#include "soapH.h"
#include "fannInterface.h"

#include <map>
#include <utility>
#include <vector>

#include "boost/filesystem/operations.hpp"
#include "boost/filesystem/path.hpp"
#include "boost/filesystem/convenience.hpp"
namespace fs = boost::filesystem;

// Path to neural network configuration and data files
static const std::string path = "net/";

///////////////////////////////////////////////////////////////

// Ignore unused argument warning with MS C++
#ifdef _MSC_VER
#pragma warning( push )
#pragma warning( disable : 4100 )
#endif /* _MSC_VER */

#if defined(_MSC_VER) && defined(DEBUG)
#define CALL_TRACE(str) \
    fprintf(stderr, "%s %s on thread %ul\n", __FUNCTION__, str, GetCurrentThreadId())
#else
#define CALL_TRACE(str)
#endif

// Define macros to report unexpected exceptions as soap faults
#define GUARD_BEGIN_NO_CONTEXT \
    try { \
    CALL_TRACE("called");

#define GUARD_BEGIN_LOOKUP_INSTANCE \
    GUARD_BEGIN_NO_CONTEXT \
    instance *inst = lookup_instance(context); \
    if (!inst) return soap_sender_fault(soap, "Context not found in fann kernel", \
        "<error xmlns=\"http://tempuri.org/\">No context instance found with the given id.</error>");

#define GUARD_BEGIN \
    GUARD_BEGIN_LOOKUP_INSTANCE \
    boost::mutex::scoped_lock lock(inst->instance_mutex); \
    inst->last_access_time = now(); \
    inst->last_access_ip = soap->ip;

#define GUARD_END \
    return SOAP_OK; \
    } \
    catch (...) \
    { \
        CALL_TRACE("faulted"); \
        return soap_sender_fault(soap, "Exception thrown in fann kernel", \
            "<error xmlns=\"http://tempuri.org/\">General exception.</error>"); \
    }

///////////////////////////////////////////////////////////////

typedef std::map<const int, instance *> context_table_t;
context_table_t context_table;
boost::mutex context_mutex;
/* Instances that have not been accessed for instance_expiry_time seconds
    are erased from the context_table in the kernel. */
const boost::xtime::xtime_sec_t purge_interval_time = 60;
boost::xtime::xtime_sec_t instance_expiry_time = 60 * 60; // Default one hour

instance *lookup_instance(int context)
{
    boost::mutex::scoped_lock lock(context_mutex);
    instance *inst = 0;
    context_table_t::iterator it = context_table.find(context);
    if (it != context_table.end())
    {
        inst = it->second;
    }
    return inst;
}

void remove_instance(int context)
{
    boost::mutex::scoped_lock lock(context_mutex);
    if (context_table.find(context) != context_table.end())
    {
        context_table.erase(context);
    }
    else
    {
        fprintf(stderr, "Context instance %d was already removed.\n", context);
    }
}

boost::xtime now(void)
{
    boost::xtime current_time;
    boost::xtime_get(&current_time, boost::TIME_UTC);
    return current_time;
}

void print_context_instances(void)
{
    const context_table_t::size_type size = context_table.size();
    fprintf(stderr, "Active contexts: %u\n", size);
    if (size > 0)
    {
        fprintf(stderr, "Context\tLast accessed by\tSeconds ago\tContent\n");
        for (context_table_t::const_iterator it = context_table.begin(); it != context_table.end(); ++it)
        {
            int context = it->first;
            instance *inst = it->second;
            fprintf(stderr, "%d\t%d.%d.%d.%d\t\t%d\t\t%s, %s, %s\n",
                context,
                (int)(inst->last_access_ip >> 24)&0xFF, (int)(inst->last_access_ip >> 16)&0xFF,
                (int)(inst->last_access_ip >> 8)&0xFF, (int)(inst->last_access_ip)&0xFF,
                (int)(now().sec - inst->last_access_time.sec),
                inst->net_name.c_str(), inst->train_name.c_str(), inst->test_name.c_str());
        }
    }
}

void purge_expired_instances(void)
{
    boost::mutex::scoped_lock lock(context_mutex);
    boost::xtime purge_time = now();

    print_context_instances();

    context_table_t::iterator it = context_table.begin();
    while (it != context_table.end())
    {
        if (it->second->last_access_time.sec < purge_time.sec - instance_expiry_time)
        {
#if defined(DEBUG)
            fprintf(stderr, "Context instance %d purged.\n", it->first);
#endif
            context_table.erase(it);       // Note: Return type in MS implementation is not standard C++ conformant
            it = context_table.begin();
        }
        else
        {
            ++it;
        }
    }
}

void purge_thread(void)
{
    for (;;)
    {
        boost::xtime invocation_time = now();
        invocation_time.sec += purge_interval_time;
        boost::thread::sleep(invocation_time);
        purge_expired_instances();
    }
}

int ns__CreateContext(struct soap *soap, void *noarg, int *context)
{
GUARD_BEGIN_NO_CONTEXT
    boost::mutex::scoped_lock lock(context_mutex);

    const context_table_t::size_type size = context_table.size();
    if (size > static_cast<context_table_t::size_type>(std::numeric_limits<int>::max()))
        return soap_sender_fault(soap, "Context limit reached in fann kernel", \
            "<error xmlns=\"http://tempuri.org/\">Limit of concurrent context instances reached.</error>"); \

    int max_context = 0;
    for (context_table_t::const_iterator it = context_table.begin(); it != context_table.end(); ++it)
    {
        if (it->first > max_context)
        {
            max_context = it->first;
        }
    }
    *context = max_context + 1;
    instance *inst = new instance;
    inst->last_access_time = now();
    inst->last_access_ip = soap->ip;
    inst->net_name = std::string("untitled.net");
    inst->train_name = std::string("untitled.train");
    inst->test_name = std::string("untitled.test");
    context_table.insert(std::make_pair(*context, inst));
#if defined(DEBUG)
    fprintf(stderr, "Context instance %d created.\n", *context);
#endif
GUARD_END
}

int ns__DestroyContext(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.destroy();
    inst->traindata.destroy_train();
    inst->testdata.destroy_train();
    remove_instance(context);
#if defined(DEBUG)
    fprintf(stderr, "Context instance %d destroyed.\n", context);
#endif
GUARD_END
}

int ns__GetContextList(struct soap *soap, void *noarg, ivector *contexts)
{
GUARD_BEGIN_NO_CONTEXT
    boost::mutex::scoped_lock lock(context_mutex);

    const context_table_t::size_type size = context_table.size();
    if (size > 0)
    {
        contexts->set_size(size);
        size_t i = 0;
        for (context_table_t::const_iterator it = context_table.begin(); it != context_table.end(); ++it)
        {
            (*contexts)[i++] = it->first;
        }
    }
GUARD_END
}

int ns__PrintContexts(struct soap *soap, void *noarg, ns__void_result *noresult)
{
GUARD_BEGIN_NO_CONTEXT
    boost::mutex::scoped_lock lock(context_mutex);
    print_context_instances();
GUARD_END
}


int ns__KeepAlive(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    // Do nothing, GUARD_BEGIN updates access time stamp.
GUARD_END
}

///////////////////////////////////////////////////////////////

int ns__GetFileList(struct soap *soap, std::string filetype, stringvector *filelist)
{
GUARD_BEGIN_NO_CONTEXT
    size_t n = 0;
    filelist->resize(n);

    fs::path full_path = fs::system_complete(fs::path(path, fs::native));
    if (fs::is_directory(full_path))
    {
        fs::directory_iterator end_it;
        for (fs::directory_iterator dir_it(full_path); dir_it != end_it; ++dir_it)
        {
            try
            {
                if (!fs::is_directory(*dir_it))
                {
                    std::string filename = dir_it->leaf();
                    std::string ext = fs::extension(*dir_it);
                    if (filetype.empty() || (ext == filetype))
                    {
                        filelist->resize(n+1);
                        (*filelist)[n] = filename;
                        ++n;
                    }
                }
            }
            catch (const std::exception &ex)
            {
                fprintf(stderr, "GetFileList %s exception %s\n", dir_it->leaf().c_str(), ex.what());
            }
        }
    }
GUARD_END
}

int ns__GetNetName(struct soap *soap, int context, std::string &filename)
{
GUARD_BEGIN
    filename = inst->net_name;
GUARD_END
}

int ns__GetTrainName(struct soap *soap, int context, std::string &filename)
{
GUARD_BEGIN
    filename = inst->train_name;
GUARD_END
}

int ns__GetTestName(struct soap *soap, int context, std::string &filename)
{
GUARD_BEGIN
    filename = inst->test_name;
GUARD_END
}

///////////////////////////////////////////////////////////////

int ns__Create(struct soap *soap, int context, enum ns__NetworkType network_type,
    ivector *layers, float connection_rate, bool *success)
{
GUARD_BEGIN
    inst->net_name = std::string("untitled.net");

    switch (network_type)
    {
    case FANN::LAYER:
        *success = inst->net.create_sparse_array(connection_rate, layers->size(),
            reinterpret_cast<const unsigned int *>(layers->__ptr));
        break;
    case FANN::SHORTCUT:
        *success = inst->net.create_shortcut_array(layers->size(),
            reinterpret_cast<const unsigned int *>(layers->__ptr));
        break;
    default:
        *success = false;
        break;
    }
GUARD_END
}

int ns__CreateFromFile(struct soap *soap, int context, std::string configuration_file, bool *success)
{
GUARD_BEGIN
    inst->net_name = configuration_file;
    *success = inst->net.create_from_file(path + configuration_file);
GUARD_END
}

int ns__RunNet(struct soap *soap, int context, dvector *input, dvector *output)
{
GUARD_BEGIN
    fann_type *outp = inst->net.run(input->__ptr);
    size_t n = inst->net.get_num_output();
    output->set_size(n);
    for (size_t i = 0; i < n; ++i)
    {
        (*output)[i] = outp[i];
    }
GUARD_END
}

int ns__RunData(struct soap *soap, int context, dmatrix *input, dmatrix *output)
{
GUARD_BEGIN
    size_t elems = input->size();
    size_t n = inst->net.get_num_output();
    output->set_size(n, elems);
    for (size_t index = 0; index < elems; ++index)
    {
        fann_type *outp = inst->net.run((*input)[index].__ptr);
        for (size_t i = 0; i < n; ++i)
        {
            (*output)[i][index] = outp[i];
        }
    }
GUARD_END
}

int ns__RunTrainData(struct soap *soap, int context, dmatrix *output)
{
GUARD_BEGIN
    unsigned int elems = inst->traindata.length_train_data();
    fann_type **data = inst->traindata.get_input();
    size_t n = inst->net.get_num_output();
    output->set_size(n, elems);
    for (size_t index = 0; index < elems; ++index)
    {
        fann_type *outp = inst->net.run(data[index]);
        for (size_t i = 0; i < n; ++i)
        {
            (*output)[i][index] = outp[i];
        }
    }
GUARD_END
}

int ns__RunTestData(struct soap *soap, int context, dmatrix *output)
{
GUARD_BEGIN
    unsigned int elems = inst->testdata.length_train_data();
    fann_type **data = inst->testdata.get_input();
    size_t n = inst->net.get_num_output();
    output->set_size(n, elems);
    for (size_t index = 0; index < elems; ++index)
    {
        fann_type *outp = inst->net.run(data[index]);
        for (size_t i = 0; i < n; ++i)
        {
            (*output)[i][index] = outp[i];
        }
    }
GUARD_END
}

int ns__RandomizeWeights(struct soap *soap, int context,
    double min_weight, double max_weight, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.randomize_weights(min_weight, max_weight);
GUARD_END
}

int ns__InitWeights(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.init_weights(inst->traindata);
GUARD_END
}

int ns__SaveNet(struct soap *soap, int context, std::string configuration_file, bool *success)
{
GUARD_BEGIN
    *success = inst->net.save(path + configuration_file);
    if (*success)
    {
        inst->net_name = configuration_file;
    }
GUARD_END
}

int ns__SaveFixed(struct soap *soap, int context, std::string configuration_file, int *fixpoint)
{
GUARD_BEGIN
    *fixpoint = inst->net.save_to_fixed(path + configuration_file);
GUARD_END
}

int ns__Train(struct soap *soap, int context, dvector *input, dvector *desired_output,
    ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.train(input->__ptr, desired_output->__ptr);
GUARD_END
}

int ns__Test(struct soap *soap, int context, dvector *input, dvector *desired_output, dvector *output)
{
GUARD_BEGIN
    fann_type *outp = inst->net.test(input->__ptr, desired_output->__ptr);
    size_t n = inst->net.get_num_output();
    output->set_size(n);
    for (size_t i = 0; i < n; ++i)
    {
        (*output)[i] = outp[i];
    }
GUARD_END
}

int ns__GetMSE(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_MSE();
GUARD_END
}

int ns__ResetMSE(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.reset_MSE();
GUARD_END
}

int ns__TrainEpoch(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.train_epoch(inst->traindata);
GUARD_END
}

int ns__TestData(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.test_data(inst->testdata);
GUARD_END
}

///////////////////////////////////////////////////////////////

// Thread specific storage used for training callback
typedef std::vector< std::pair<unsigned int, float> > mse_t;
boost::thread_specific_ptr<mse_t> tss_data;
boost::thread_specific_ptr<instance> tss_inst;

// Training callback used by TrainOnData and TrainOnFile
int train_callback(FANN::neural_net &net, FANN::training_data &train,
    unsigned int max_epochs, unsigned int epochs_between_reports,
    float desired_error, unsigned int epochs, void *user_data)
{
    try
    {
        instance *inst = tss_inst.get();
        mse_t *result = tss_data.get();
        result->push_back(std::make_pair(epochs, net.get_MSE()));
        inst->last_access_time = now(); // In case the training is taking a very long time
        if (inst->stop_training)
        {
            return -1;
        }
        else
        {
            return 0;
        }
    }
    catch (...)
    {
        fprintf(stderr, "Exception in training callback\n");
        return -1;
    }
}

int ns__StopTraining(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN_LOOKUP_INSTANCE
    // Not protected by the instance mutex
    inst->stop_training = true;
GUARD_END
}

int ns__TrainOnData(struct soap *soap, int context, unsigned int max_epochs,
    unsigned int epochs_between_reports, float desired_error, dmatrix *result)
{
GUARD_BEGIN
    mse_t mse;
    tss_data.reset(&mse); // initialize the thread's storage
    inst->stop_training = false;
    tss_inst.reset(inst);

    inst->net.set_callback(train_callback, NULL);
    inst->net.train_on_data(inst->traindata, max_epochs,
        epochs_between_reports, desired_error);

    result->set_size(mse.size(), 2);
    int i = 0;
    for (mse_t::const_iterator it = mse.begin(); it != mse.end(); ++it)
    {
        (*result)[i][0] = it->first;
        (*result)[i][1] = it->second;
        ++i;
    }
    inst->last_access_time = now(); // In case the training took a very long time
    tss_inst.release();   // Release tread local storage to avoid cleanup problems
    tss_data.release();
GUARD_END
}

int ns__TrainOnFile(struct soap *soap, int context, std::string filename, unsigned int max_epochs,
    unsigned int epochs_between_reports, float desired_error, dmatrix *result)
{
GUARD_BEGIN
    mse_t mse;
    tss_data.reset(&mse); // initialize the thread's storage
    inst->stop_training = false;
    tss_inst.reset(inst);

    inst->net.set_callback(train_callback, NULL);
    inst->net.train_on_file(path + filename, max_epochs,
        epochs_between_reports, desired_error);

    result->set_size(mse.size(), 2);
    int i = 0;
    for (mse_t::const_iterator it = mse.begin(); it != mse.end(); ++it)
    {
        (*result)[i][0] = it->first;
        (*result)[i][1] = it->second;
        ++i;
    }
    inst->last_access_time = now(); // In case the training took a very long time
    tss_inst.release();   // Release tread local storage to avoid cleanup problems
    tss_data.release();
GUARD_END
}

int ns__PrintParameters(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.print_parameters();
GUARD_END
}

int ns__PrintConnections(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.print_connections();
GUARD_END
}

/* Cascade training and properties */

int ns__CascadeTrainOnData(struct soap *soap, int context, unsigned int maxneurons,
    unsigned int neuronsbetweenreports, float desirederror, dmatrix *result)
{
GUARD_BEGIN
    mse_t mse;
    tss_data.reset(&mse); // initialize the thread's storage
    inst->stop_training = false;
    tss_inst.reset(inst);

    inst->net.set_callback(train_callback, NULL);
    inst->net.cascadetrain_on_data(inst->traindata, maxneurons,
        neuronsbetweenreports, desirederror);

    result->set_size(mse.size(), 2);
    int i = 0;
    for (mse_t::const_iterator it = mse.begin(); it != mse.end(); ++it)
    {
        (*result)[i][0] = it->first;
        (*result)[i][1] = it->second;
        ++i;
    }
    inst->last_access_time = now(); // In case the training took a very long time
    tss_inst.release();   // Release tread local storage to avoid cleanup problems
    tss_data.release();
GUARD_END
}

int ns__CascadeTrainOnFile(struct soap *soap, int context, std::string filename, unsigned int maxneurons,
    unsigned int neuronsbetweenreports, float desirederror, dmatrix *result)
{
GUARD_BEGIN
    mse_t mse;
    tss_data.reset(&mse); // initialize the thread's storage
    inst->stop_training = false;
    tss_inst.reset(inst);

    inst->net.set_callback(train_callback, NULL);
    inst->net.cascadetrain_on_file(path + filename, maxneurons,
        neuronsbetweenreports, desirederror);

    result->set_size(mse.size(), 2);
    int i = 0;
    for (mse_t::const_iterator it = mse.begin(); it != mse.end(); ++it)
    {
        (*result)[i][0] = it->first;
        (*result)[i][1] = it->second;
        ++i;
    }
    inst->last_access_time = now(); // In case the training took a very long time
    tss_inst.release();   // Release tread local storage to avoid cleanup problems
    tss_data.release();
GUARD_END
}

int ns__GetCascadeOutputChangeFraction(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_cascade_output_change_fraction();
GUARD_END
}

int ns__SetCascadeOutputChangeFraction(struct soap *soap, int context, float property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_output_change_fraction(property);
GUARD_END
}

int ns__GetCascadeOutputStagnationEpochs(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_cascade_output_stagnation_epochs();
GUARD_END
}

int ns__SetCascadeOutputStagnationEpochs(struct soap *soap, int context, unsigned int property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_output_stagnation_epochs(property);
GUARD_END
}

int ns__GetCascadeCandidateChangeFraction(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_cascade_candidate_change_fraction();
GUARD_END
}

int ns__SetCascadeCandidateChangeFraction(struct soap *soap, int context, float property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_candidate_change_fraction(property);
GUARD_END
}

int ns__GetCascadeCandidateStagnationEpochs(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_cascade_candidate_stagnation_epochs();
GUARD_END
}

int ns__SetCascadeCandidateStagnationEpochs(struct soap *soap, int context, unsigned int property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_candidate_stagnation_epochs(property);
GUARD_END
}

int ns__GetCascadeWeightMultiplier(struct soap *soap, int context, double *result)
{
GUARD_BEGIN
    *result = inst->net.get_cascade_weight_multiplier();
GUARD_END
}

int ns__SetCascadeWeightMultiplier(struct soap *soap, int context, double property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_weight_multiplier(property);
GUARD_END
}

int ns__GetCascadeCandidateLimit(struct soap *soap, int context, double *result)
{
GUARD_BEGIN
    *result = inst->net.get_cascade_candidate_limit();
GUARD_END
}

int ns__SetCascadeCandidateLimit(struct soap *soap, int context, double property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_candidate_limit(property);
GUARD_END
}

int ns__GetCascadeCandidateMaximumOutputEpochs(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_cascade_max_out_epochs();
GUARD_END
}

int ns__SetCascadeCandidateMaximumOutputEpochs(struct soap *soap, int context, unsigned int property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_max_out_epochs(property);
GUARD_END
}

int ns__GetCascadeCandidateMaximumCandidateEpochs(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_cascade_max_cand_epochs();
GUARD_END
}

int ns__SetCascadeCandidateMaximumCandidateEpochs(struct soap *soap, int context, unsigned int property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_max_cand_epochs(property);
GUARD_END
}

int ns__GetCascadeCandidateNumCandidates(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_cascade_num_candidates();
GUARD_END
}

int ns__GetCascadeCandidateNumCandidateGroups(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_cascade_num_candidate_groups();
GUARD_END
}

int ns__SetCascadeCandidateNumCandidateGroups(struct soap *soap, int context, unsigned int property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_num_candidate_groups(property);
GUARD_END
}

int ns__GetCascadeActivationFunctions(struct soap *soap, int context, afvector *output)
{
GUARD_BEGIN
    FANN::activation_function_enum *outp = inst->net.get_cascade_activation_functions();
    size_t n = inst->net.get_cascade_activation_functions_count();
    output->set_size(n);
    for (size_t i = 0; i < n; ++i)
    {
        (*output)[i] = static_cast<enum ns__ActivationFunction>(outp[i]);
    }
GUARD_END
}

int ns__SetCascadeActivationFunctions(struct soap *soap, int context, afvector *input, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_activation_functions(
        reinterpret_cast<FANN::activation_function_enum *>(input->__ptr), input->size());
GUARD_END
}

int ns__GetCascadeActivationSteepnesses(struct soap *soap, int context, dvector *output)
{
GUARD_BEGIN
    fann_type *outp = inst->net.get_cascade_activation_steepnesses();
    size_t n = inst->net.get_cascade_activation_steepnesses_count();
    output->set_size(n);
    for (size_t i = 0; i < n; ++i)
    {
        (*output)[i] = outp[i];
    }
GUARD_END
}

int ns__SetCascadeActivationSteepnesses(struct soap *soap, int context, dvector *input, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_cascade_activation_steepnesses(input->__ptr, input->size());
GUARD_END
}

/***************************************************************************************************/

int ns__GetTrainingAlgorithm(struct soap *soap, int context,
    enum ns__TrainingAlgorithm *result)
{
GUARD_BEGIN
    *result = static_cast<enum ns__TrainingAlgorithm>(inst->net.get_training_algorithm());
GUARD_END
}

int ns__SetTrainingAlgorithm(struct soap *soap, int context,
    enum ns__TrainingAlgorithm property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_training_algorithm(
        static_cast<enum FANN::training_algorithm_enum>(property));
GUARD_END
}

int ns__GetLearningRate(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_learning_rate();
GUARD_END
}

int ns__SetLearningRate(struct soap *soap, int context, float property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_learning_rate(property);
GUARD_END
}

int ns__GetActivationFunction(struct soap *soap, int context, int layer, int neuron, enum ns__ActivationFunction *result)
{
GUARD_BEGIN
    *result = static_cast<enum ns__ActivationFunction>
        (inst->net.get_activation_function(layer, neuron));
GUARD_END
}

int ns__SetActivationFunction(struct soap *soap, int context,
    enum ns__ActivationFunction property, int layer, int neuron, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_activation_function(
        static_cast<enum FANN::activation_function_enum>(property), layer, neuron);
GUARD_END
}

int ns__SetActivationFunctionLayer(struct soap *soap, int context,
    enum ns__ActivationFunction property, int layer, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_activation_function_layer(
        static_cast<enum FANN::activation_function_enum>(property), layer);
GUARD_END
}

int ns__SetActivationFunctionHidden(struct soap *soap, int context,
    enum ns__ActivationFunction property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_activation_function_hidden(
        static_cast<enum FANN::activation_function_enum>(property));
GUARD_END
}

int ns__SetActivationFunctionOutput(struct soap *soap, int context,
    enum ns__ActivationFunction property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_activation_function_output(
        static_cast<enum FANN::activation_function_enum>(property));
GUARD_END
}

int ns__GetActivationSteepness(struct soap *soap, int context, int layer, int neuron, double *result)
{
GUARD_BEGIN
    *result = inst->net.get_activation_steepness(layer, neuron);
GUARD_END
}

int ns__SetActivationSteepness(struct soap *soap, int context,
    double property, int layer, int neuron, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_activation_steepness(property, layer, neuron);
GUARD_END
}

int ns__SetActivationSteepnessLayer(struct soap *soap, int context,
    double property, int layer, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_activation_steepness_layer(property, layer);
GUARD_END
}

int ns__SetActivationSteepnessHidden(struct soap *soap, int context,
    double property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_activation_steepness_hidden(property);
GUARD_END
}

int ns__SetActivationSteepnessOutput(struct soap *soap, int context,
    double property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_activation_steepness_output(property);
GUARD_END
}

int ns__GetErrorFunction(struct soap *soap, int context, enum ns__ErrorFunction *result)
{
GUARD_BEGIN
    *result = static_cast<enum ns__ErrorFunction>(inst->net.get_train_error_function());
GUARD_END
}

int ns__SetErrorFunction(struct soap *soap, int context,
    enum ns__ErrorFunction property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_train_error_function(
        static_cast<enum FANN::error_function_enum>(property));
GUARD_END
}

int ns__GetQuickpropDecay(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_quickprop_decay();
GUARD_END
}

int ns__SetQuickpropDecay(struct soap *soap, int context, float property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_quickprop_decay(property);
GUARD_END
}

int ns__GetQuickpropMu(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_quickprop_mu();
GUARD_END
}

int ns__SetQuickpropMu(struct soap *soap, int context, float property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_quickprop_mu(property);
GUARD_END
}

int ns__GetRpropIncreaseFactor(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_rprop_increase_factor();
GUARD_END
}

int ns__SetRpropIncreaseFactor(struct soap *soap, int context, float property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_rprop_increase_factor(property);
GUARD_END
}

int ns__GetRpropDecreaseFactor(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_rprop_decrease_factor();
GUARD_END
}

int ns__SetRpropDecreaseFactor(struct soap *soap, int context, float property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_rprop_decrease_factor(property);
GUARD_END
}

int ns__GetRpropDeltaMin(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_rprop_delta_min();
GUARD_END
}

int ns__SetRpropDeltaMin(struct soap *soap, int context, float property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_rprop_delta_min(property);
GUARD_END
}

int ns__GetRpropDeltaMax(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_rprop_delta_max();
GUARD_END
}

int ns__SetRpropDeltaMax(struct soap *soap, int context, float property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_rprop_delta_max(property);
GUARD_END
}

int ns__GetNumInput(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_num_input();
GUARD_END
}

int ns__GetNumOutput(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_num_output();
GUARD_END
}

int ns__GetTotalNeurons(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_total_neurons();
GUARD_END
}

int ns__GetTotalConnections(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_total_connections();
GUARD_END
}

int ns__GetNetworkType(struct soap *soap, int context, enum ns__NetworkType *result)
{
GUARD_BEGIN
    *result = static_cast<enum ns__NetworkType>(inst->net.get_network_type());
GUARD_END
}

int ns__GetConnectionRate(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_connection_rate();
GUARD_END
}

int ns__GetNumLayers(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_num_layers();
GUARD_END
}

int ns__GetLayers(struct soap *soap, int context, ivector *result)
{
GUARD_BEGIN
    unsigned int num_layers = inst->net.get_num_layers();
    unsigned int *layers = new unsigned int[num_layers];
    if (layers != NULL)
    {
        inst->net.get_layer_array(layers);
        result->set_size(num_layers);
        for (size_t i = 0; i < num_layers; ++i)
        {
            (*result)[i] = layers[i];
        }
        delete(layers);
    }
GUARD_END
}

int ns__GetBias(struct soap *soap, int context, ivector *result)
{
GUARD_BEGIN
    unsigned int num_bias = inst->net.get_num_layers();
    unsigned int *bias = new unsigned int[num_bias];
    if (bias != NULL)
    {
        inst->net.get_bias_array(bias);
        result->set_size(num_bias);
        for (size_t i = 0; i < num_bias; ++i)
        {
            (*result)[i] = bias[i];
        }
        delete(bias);
    }
GUARD_END
}

int ns__GetConnections(struct soap *soap, int context, dmatrix *result)
{
GUARD_BEGIN
    unsigned int num_connections = inst->net.get_total_connections();
    FANN::connection *connections = new FANN::connection[num_connections];
    if (connections != NULL)
    {
        inst->net.get_connection_array(connections);
        result->set_size(num_connections, 3);
        for (unsigned int index = 0; index < num_connections; index++)
        {
            (*result)[index][0] = connections[index].from_neuron;
            (*result)[index][1] = connections[index].to_neuron;
            (*result)[index][2] = connections[index].weight;
        }
        delete(connections);
    }
GUARD_END
}

int ns__SetWeight(struct soap *soap, int context, unsigned int from_neuron,
    unsigned int to_neuron, double weight, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_weight(from_neuron, to_neuron, weight);
GUARD_END
}

int ns__SetWeights(struct soap *soap, int context, dmatrix *connections, ns__void_result *noresult)
{
GUARD_BEGIN
    for (int index = 0; index < connections->size(); index++)
    {
        unsigned int from_neuron = static_cast<unsigned int>((*connections)[index][0]);
        unsigned int to_neuron = static_cast<unsigned int>((*connections)[index][1]);
        double weight = (*connections)[index][2];
        inst->net.set_weight(from_neuron, to_neuron, weight);
    }
GUARD_END
}


int ns__GetLearningMomentum(struct soap *soap, int context, float *result)
{
GUARD_BEGIN
    *result = inst->net.get_learning_momentum();
GUARD_END
}

int ns__SetLearningMomentum(struct soap *soap, int context, float property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_learning_momentum(property);
GUARD_END
}

int ns__GetStopFunction(struct soap *soap, int context, enum ns__StopFunction *result)
{
GUARD_BEGIN
    *result = static_cast<enum ns__StopFunction>(inst->net.get_train_stop_function());
GUARD_END
}

int ns__SetStopFunction(struct soap *soap, int context, enum ns__StopFunction property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_train_stop_function(
        static_cast<enum FANN::stop_function_enum>(property));
GUARD_END
}

int ns__GetBitFailLimit(struct soap *soap, int context, double *result)
{
GUARD_BEGIN
    *result = inst->net.get_bit_fail_limit();
GUARD_END
}

int ns__SetBitFailLimit(struct soap *soap, int context, double property, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.set_bit_fail_limit(property);
GUARD_END
}

int ns__GetBitFail(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->net.get_bit_fail();
GUARD_END
}

/*********************************************************************/

int ns__ScaleTrain(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.scale_train(inst->traindata);
GUARD_END
}

int ns__DescaleTrain(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.descale_train(inst->traindata);
GUARD_END
}

int ns__ScaleTest(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.scale_train(inst->testdata);
GUARD_END
}

int ns__DescaleTest(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.descale_train(inst->testdata);
GUARD_END
}

int ns__SetInputScalingParams(struct soap *soap, int context, float newinputmin, float newinputmax, bool *success)
{
GUARD_BEGIN
    *success = inst->net.set_input_scaling_params(inst->traindata, newinputmin, newinputmax);
GUARD_END
}

int ns__SetOutputScalingParams(struct soap *soap, int context, float newoutputmin, float newoutputmax, bool *success)
{
GUARD_BEGIN
    *success = inst->net.set_output_scaling_params(inst->traindata, newoutputmin, newoutputmax);
GUARD_END
}

int ns__SetScalingParams(struct soap *soap, int context, float newinputmin, float newinputmax, float newoutputmin, float newoutputmax, bool *success)
{
GUARD_BEGIN
    *success = inst->net.set_scaling_params(inst->traindata, newinputmin, newinputmax, newoutputmin, newoutputmax);
GUARD_END
}

int ns__ClearScalingParams(struct soap *soap, int context, bool *success)
{
GUARD_BEGIN
    *success = inst->net.clear_scaling_params();
GUARD_END
}

int ns__ScaleInput(struct soap *soap, int context, dvector *inputvector, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.scale_input(inputvector->__ptr);
GUARD_END
}

int ns__ScaleOutput(struct soap *soap, int context, dvector *outputvector, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.scale_output(outputvector->__ptr);
GUARD_END
}

int ns__DescaleInput(struct soap *soap, int context, dvector *inputvector, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.descale_input(inputvector->__ptr);
GUARD_END
}

int ns__DescaleOutput(struct soap *soap, int context, dvector *outputvector, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->net.descale_output(outputvector->__ptr);
GUARD_END
}

/*********************************************************************/

int ns__ReadTrainFromFile(struct soap *soap, int context, std::string data_file, bool *success)
{
GUARD_BEGIN
    inst->train_name = data_file;
    *success = inst->traindata.read_train_from_file(path + data_file);
GUARD_END
}

int ns__SaveTrain(struct soap *soap, int context, std::string data_file, bool *success)
{
GUARD_BEGIN
    *success = inst->traindata.save_train(path + data_file);
    if (*success)
    {
        inst->train_name = data_file;
    }
GUARD_END
}

int ns__SaveTrainToFixed(struct soap *soap, int context, std::string data_file,
    unsigned int decimal_point, bool *success)
{
GUARD_BEGIN
    *success = inst->traindata.save_train_to_fixed(path + data_file, decimal_point);
GUARD_END
}

int ns__ShuffleTrainData(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->traindata.shuffle_train_data();
GUARD_END
}

int ns__MergeTestIntoTrainData(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->traindata.merge_train_data(inst->testdata);
GUARD_END
}

int ns__GetTrainNumData(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->traindata.length_train_data();
GUARD_END
}

int ns__GetTrainNumInput(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->traindata.num_input_train_data();
GUARD_END
}

int ns__GetTrainNumOutput(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->traindata.num_output_train_data();
GUARD_END
}

int ns__GetTrainInput(struct soap *soap, int context, dmatrix *result)
{
GUARD_BEGIN
    unsigned int rows = inst->traindata.length_train_data();
    unsigned int cols = inst->traindata.num_input_train_data();
    fann_type **data = inst->traindata.get_input();
    result->set_size(cols, rows);
    for (unsigned int i = 0; i < rows; ++i)
    {
        for (unsigned int j = 0; j < cols; ++j)
        {
            (*result)[j][i] = data[i][j];
        }
    }
GUARD_END
}

int ns__GetTrainOutput(struct soap *soap, int context, dmatrix *result)
{
GUARD_BEGIN
    unsigned int rows = inst->traindata.length_train_data();
    unsigned int cols = inst->traindata.num_output_train_data();
    fann_type **data = inst->traindata.get_output();
    result->set_size(cols, rows);
    for (unsigned int i = 0; i < rows; ++i)
    {
        for (unsigned int j = 0; j < cols; ++j)
        {
            (*result)[j][i] = data[i][j];
        }
    }
GUARD_END
}

int ns__SetTrainData(struct soap *soap, int context,
    dmatrix *input, dmatrix *output, ns__void_result *noresult)
{
GUARD_BEGIN
    if (input->size() != output->size())
    {
        return soap_sender_fault(soap, "Input/output data mismatch", \
            "<error xmlns=\"http://tempuri.org/\">The number of input and output elements does not match.</error>");
    }
    if (input->size() < 1)
    {
        inst->traindata.destroy_train();
    }
    else
    {
        unsigned int num_data = input->size();
        unsigned int num_input = (*input)[0].size();
        unsigned int num_output = (*output)[0].size();
        fann_type **indata = new fann_type *[num_data];
        fann_type **outdata = new fann_type *[num_data];

        for (unsigned int i = 0; i < num_data; ++i)
        {
            indata[i] = (*input)[i].__ptr;
            outdata[i] = (*output)[i].__ptr;
        }
        inst->traindata.set_train_data(num_data, num_input, indata, num_output, outdata);
        delete(indata);
        delete(outdata);
    }
GUARD_END
}


int ns__ScaleTrainData(struct soap *soap, int context, double minimum, double maximum, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->traindata.scale_train_data(minimum, maximum);
GUARD_END
}

int ns__ScaleTrainInput(struct soap *soap, int context, double minimum, double maximum, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->traindata.scale_input_train_data(minimum, maximum);
GUARD_END
}

int ns__ScaleTrainOutput(struct soap *soap, int context, double minimum, double maximum, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->traindata.scale_output_train_data(minimum, maximum);
GUARD_END
}

int ns__SubsetTrainData(struct soap *soap, int context, unsigned int position, unsigned int length, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->traindata.subset_train_data(position, length);
GUARD_END
}


int ns__ReadTestFromFile(struct soap *soap, int context, std::string data_file, bool *success)
{
GUARD_BEGIN
    inst->test_name = data_file;
    *success = inst->testdata.read_train_from_file(path + data_file);
GUARD_END
}

int ns__SaveTest(struct soap *soap, int context, std::string data_file, bool *success)
{
GUARD_BEGIN
    *success = inst->testdata.save_train(path + data_file);
    if (*success)
    {
        inst->test_name = data_file;
    }
GUARD_END
}

int ns__SaveTestToFixed(struct soap *soap, int context, std::string data_file,
    unsigned int decimal_point, bool *success)
{
GUARD_BEGIN
    *success = inst->testdata.save_train_to_fixed(path + data_file, decimal_point);
GUARD_END
}

int ns__ShuffleTestData(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->testdata.shuffle_train_data();
GUARD_END
}

int ns__MergeTrainIntoTestData(struct soap *soap, int context, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->testdata.merge_train_data(inst->traindata);
GUARD_END
}

int ns__GetTestNumData(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->testdata.length_train_data();
GUARD_END
}

int ns__GetTestNumInput(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->testdata.num_input_train_data();
GUARD_END
}

int ns__GetTestNumOutput(struct soap *soap, int context, unsigned int *result)
{
GUARD_BEGIN
    *result = inst->testdata.num_output_train_data();
GUARD_END
}

int ns__GetTestInput(struct soap *soap, int context, dmatrix *result)
{
GUARD_BEGIN
    unsigned int rows = inst->testdata.length_train_data();
    unsigned int cols = inst->testdata.num_input_train_data();
    fann_type **data = inst->testdata.get_input();
    result->set_size(cols, rows);
    for (unsigned int i = 0; i < rows; ++i)
    {
        for (unsigned int j = 0; j < cols; ++j)
        {
            (*result)[j][i] = data[i][j];
        }
    }
GUARD_END
}

int ns__GetTestOutput(struct soap *soap, int context, dmatrix *result)
{
GUARD_BEGIN
    unsigned int rows = inst->testdata.length_train_data();
    unsigned int cols = inst->testdata.num_output_train_data();
    fann_type **data = inst->testdata.get_output();
    result->set_size(cols, rows);
    for (unsigned int i = 0; i < rows; ++i)
    {
        for (unsigned int j = 0; j < cols; ++j)
        {
            (*result)[j][i] = data[i][j];
        }
    }
GUARD_END
}

int ns__SetTestData(struct soap *soap, int context,
    dmatrix *input, dmatrix *output, ns__void_result *noresult)
{
GUARD_BEGIN
    if (input->size() != output->size())
    {
        return soap_sender_fault(soap, "Input/output data mismatch", \
            "<error xmlns=\"http://tempuri.org/\">The number of input and output elements does not match.</error>");
    }
    if (input->size() < 1)
    {
        inst->testdata.destroy_train();
    }
    else
    {
        unsigned int num_data = input->size();
        unsigned int num_input = (*input)[0].size();
        unsigned int num_output = (*output)[0].size();
        fann_type **indata = new fann_type *[num_data];
        fann_type **outdata = new fann_type *[num_data];

        for (unsigned int i = 0; i < num_data; ++i)
        {
            indata[i] = (*input)[i].__ptr;
            outdata[i] = (*output)[i].__ptr;
        }
        inst->testdata.set_train_data(num_data, num_input, indata, num_output, outdata);
        delete(indata);
        delete(outdata);
    }
GUARD_END
}


int ns__ScaleTestData(struct soap *soap, int context, double minimum, double maximum, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->testdata.scale_train_data(minimum, maximum);
GUARD_END
}

int ns__ScaleTestInput(struct soap *soap, int context, double minimum, double maximum, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->testdata.scale_input_train_data(minimum, maximum);
GUARD_END
}

int ns__ScaleTestOutput(struct soap *soap, int context, double minimum, double maximum, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->testdata.scale_output_train_data(minimum, maximum);
GUARD_END
}

int ns__SubsetTestData(struct soap *soap, int context, unsigned int position, unsigned int length, ns__void_result *noresult)
{
GUARD_BEGIN
    inst->testdata.subset_train_data(position, length);
GUARD_END
}


#ifdef _MSC_VER
#pragma warning( pop )
#endif /* _MSC_VER */

////////////////////////////////////////////////////////////
