// Copyright 2005/2548 written by freegoldbar at yahoo dot com.
// All rights reserved. For private and educational use.
// Not for sale, commercial or military use.
// No warranties given. No suitability for any purpose.

#include "soapH.h"
#include "fannKernel.nsmap"
#include "fannInterface.h"
#include <ios>
#include <iostream>
#include <fstream>
#include <iterator>
#include <string>
#include <vector>
#include <boost/algorithm/string.hpp>

// Disable warning in lexical_cast implementation
#ifdef _MSC_VER
#pragma warning(disable: 4701)
#endif
#include <boost/lexical_cast.hpp>
#ifdef _MSC_VER
#pragma warning(default: 4701)
#endif
#include <boost/program_options.hpp>
namespace options = boost::program_options;

#define VERSION "2.1.0"
#define WSDL_FILE_NAME "fannKernel.wsdl"
#define DEFAULT_HOST "127.0.0.1"
#define DEFAULT_PORT 2718 // First four digits from exponential constant
#define DEFAULT_EXPIRY 60 // Default one hour before cleaning up unused context instances
#define BACKLOG 100 // Max. request backlog
#define SERVER_TIMEOUT 0 // Timeout for server inactivity
#define TIMEOUT 0 // Common timeout for send and receive stalls
#define HOST_NAME_MAX_LENGTH 512 // Max. length of a host name

static char host_name[HOST_NAME_MAX_LENGTH] = DEFAULT_HOST; // Initialized before threading starts up
static int port = DEFAULT_PORT; // set default port

static const std::string web_path = "fannExplorer/";

bool parse_program_options(int argc, char *argv[])
{
    try
    {
        int expiry;

        // Options that are only allowed on the command line
        options::options_description generic("Generic options");
        generic.add_options()
            ("version,v", "display version string")
            ("help,?", "display this help message")
            ;

        // Options that are only allowed on both the command line and in a configuration file
        options::options_description config("Configuration");
        config.add_options()
            ("port,p", options::value<int>(&port)->default_value(DEFAULT_PORT),
                "ip port used for serving wsdl and soap requests")
            ("expiry,e", options::value<int>(&expiry)->default_value(DEFAULT_EXPIRY),
                "expiry time for unused context instances in minutes")
            ;

        // Hidden options are allowed both on the command line and in a configuration file
        options::options_description hidden("Hidden options");
        hidden.add_options()
            ("input-file", options::value< std::vector<std::string> >(), "input file")
            ;

        options::options_description cmdline_options;
        cmdline_options.add(generic).add(config).add(hidden);

        options::options_description config_file_options;
        config_file_options.add(config).add(hidden);

        options::options_description visible("Allowed options");
        visible.add(generic).add(config);
        
        options::positional_options_description p;
        p.add("input-file", -1);
        
        options::variables_map vm;
        store(options::command_line_parser(argc, argv).options(cmdline_options).positional(p).run(), vm);

        std::ifstream ifs("fannKernel.cfg");
        store(parse_config_file(ifs, config_file_options), vm);
        notify(vm);
    
        if (vm.count("help")) {
            std::cout << "fannKernel, version " << VERSION << std::endl << std::endl;
            std::cout << "fast artificial neural network calculation kernel with soap interface" << std::endl;
            std::cout << "options can be set on the command line or in a fannKernel.cfg file" << std::endl << std::endl;
            std::cout << visible << std::endl;
            return false;
        }

        if (vm.count("version")) {
            std::cout << "fannKernel, version " << VERSION << std::endl;
            return false;
        }
        instance_expiry_time = expiry * 60; // Convert from minutes to seconds
        return true;
    }
    catch (std::exception& e)
    {
        std::cerr << "Command line exception: " << e.what() << std::endl;
        return false;
    }
}


// Serves wsdl - expanded to serve HTML and SWF files
int http_get(struct soap *soap)
{
    std::string path(soap->path);
#if defined(DEBUG)
    std::cout << "HTTP GET Request: " << path << std::endl;
#endif
    std::string::size_type pos = path.find_last_of('?');
    if ((pos != std::string::npos) && (boost::algorithm::to_lower_copy(path.substr(pos)).compare("?wsdl") == 0))
    {
        std::cout << "Serving WSDL Item" << std::endl;
        FILE *fd = fopen(WSDL_FILE_NAME, "rb"); // open WSDL file to copy
        if (!fd)
            return 404; // return HTTP not found error

        std::string wsdl;
        for (;;)
        {
            size_t size_read = fread(soap->tmpbuf, 1, sizeof(soap->tmpbuf), fd);
            if (!size_read)
                break;
            wsdl.append(soap->tmpbuf, size_read);
        }
        fclose(fd);

        std::string pattern = std::string(DEFAULT_HOST) + ":" + boost::lexical_cast<std::string>(DEFAULT_PORT);
        std::string replace = std::string(host_name) + ":" + boost::lexical_cast<std::string>(port);
        boost::algorithm::replace_all(wsdl, pattern, replace);

        soap->http_content = "text/xml"; // HTTP header with text/xml content
        soap_response(soap, SOAP_FILE);
        int ret = soap_send_raw(soap, wsdl.data(), wsdl.length());
        if (ret) {
            std::cerr << "WSDL HTTP GET failed" << std::endl;
        }
#if defined(DEBUG)
        else {
            std::cout << "WSDL HTTP GET successful" << std::endl;
        }
#endif
        soap_end_send(soap);
        return SOAP_OK;
    }
    else
    {
        std::string::size_type pos = path.find_last_of('/');
        if (pos != std::string::npos)
        {
            std::string web_item = web_path + path.substr(pos+1);
            std::cout << "Serving Web Item: " << web_item << std::endl;
            FILE *fd = fopen(web_item.c_str(), "rb"); // open file to copy
            if (!fd)
                return 404; // return HTTP not found error

            std::string content;
            for (;;)
            {
                size_t size_read = fread(soap->tmpbuf, 1, sizeof(soap->tmpbuf), fd);
                if (!size_read)
                    break;
                content.append(soap->tmpbuf, size_read);
            }
            fclose(fd);

            std::string::size_type pos_extension = path.find_last_of('.');
            if (boost::algorithm::to_lower_copy(path.substr(pos)).compare(".swf") == 0)
                soap->http_content = "image/swf"; // HTTP header with flash content
            else
                soap->http_content = "text/html"; // HTTP header with text/html content
            soap_response(soap, SOAP_FILE);
            int ret = soap_send_raw(soap, content.data(), content.length());
            if (ret) {
                std::cerr << "HTTP GET failed" << std::endl;
            }
#if defined(DEBUG)
            else {
                std::cout << "HTTP GET successful" << std::endl;
            }
#endif
            soap_end_send(soap);
            return SOAP_OK;
        }
        else
        {
            std::cout << "Received unsupported request: " << path << std::endl;
            return SOAP_GET_METHOD;
        }
    }
}

void process_request(void *psoap)
{
    try
    {
        struct soap *soap = static_cast<struct soap*>(psoap);
        soap->recv_timeout = TIMEOUT; // Timeout after stall on recv 
        soap->send_timeout = TIMEOUT; // Timeout after stall on send 
        soap->http_content = "text/xml"; // HTTP header with text/xml content

        soap_serve(soap);
#if defined(DEBUG)
        std::cout << "Client thread exiting" << std::endl;
#endif

        soap_destroy(soap);
        soap_end(soap);
        soap_done(soap);
        free(soap);
    }
    catch (...)
    {
        std::cerr << "Generel exception in soap request processing" << std::endl;
    }
}

struct thread_adapt
{
    thread_adapt(void (*func)(void*), void* param) : _func(func), _param(param) {}
    int operator()() const
    {
        _func(_param);
        return 0;
    }

    void (*_func)(void*);
    void* _param;
};

class thread_adapter
{
public:
    thread_adapter(void (*func)(void*), void* param) : _func(func), _param(param) {}
    void operator()() const { _func(_param); }
private:
    void (*_func)(void*);
    void* _param;
};

int main(int argc, char **argv)
{
    try
    {
        std::ios::sync_with_stdio();
        if (parse_program_options(argc, argv))
        {
            std::cout << "fannKernel " << VERSION << std::endl;
            std::cout << "Listening for clients on TCP/IP port " << port << std::endl;
            std::cout << "Automatic cleanup of unused data after " << (int)(instance_expiry_time/60) << " minutes" << std::endl;
            std::cout << "Use --help to list startup options" << std::endl;

            struct soap soap;
            soap_init2(&soap, SOAP_IO_KEEPALIVE, SOAP_IO_KEEPALIVE | SOAP_IO_CHUNK);
            // set maximum number of calls per keep-alive session
            soap.max_keep_alive = std::numeric_limits<unsigned int>::max();
            soap.accept_timeout = SERVER_TIMEOUT; // server timeout after inactivity
            soap.fget = http_get;

            int bind_master = soap_bind(&soap, NULL, port, BACKLOG);
            if (bind_master < 0)
            {
                soap_print_fault(&soap, stderr);
                return (1);
            }
            std::cout << "Socket connection " << bind_master << " successful" << std::endl;

            int status = gethostname(host_name, HOST_NAME_MAX_LENGTH);
            if (status == 0)
            {
                std::string browser = static_cast<std::string>("http://") +
                    static_cast<std::string>(host_name) + static_cast<std::string>(":") +
                    boost::lexical_cast<std::string>(port) + static_cast<std::string>("/fannExplorer.html");
                std::cout << std::endl << "To run fannExplorer enter in browser: " << browser << std::endl;
                std::cout << "To get fannKernel WSDL/SOAP service definition use: " <<
                    "http://" << host_name << ":" << port << "/?wsdl" << std::endl;

#ifdef _WIN32 // TODO
//                std::cout << std::endl << std::endl;
//                std::cout << static_cast<std::string>("\"C:\\Program Files\\Internet Explorer\\iexplore.exe\" \"") <<
//                    browser << static_cast<std::string>("\"");
//                std::cout << std:endl;
//                system((static_cast<std::string>("\"C:\\Program Files\\Internet Explorer\\iexplore.exe\" \"") + // TODO
//                    browser + static_cast<std::string>("\"")).c_str());
#endif
            }
            else
            {
                std::cerr << "Socket gethostname error " << status << std::endl;
                return (1);
            }

            boost::thread_group threads;
            threads.create_thread(purge_thread);
            for (;;)
            { 
#ifndef _WIN32
                soap.socket_flags = MSG_NOSIGNAL; // use this on Unix/Linux to prevent sigpipe
#endif
                int client = soap_accept(&soap); 
                if (client < 0)
                {
                    // Assume timeout is long enough for threads to complete serving requests
                    if (soap.errnum)
                        soap_print_fault(&soap, stderr);
                    else
                        std::cerr << "Server timed out" << std::endl;
                    break;
                }
#if defined(DEBUG)
                std::cout << "Accepts socket connection " << client << " from IP " <<
                    ((soap.ip >> 24)&0xFF) << "." << ((soap.ip >> 16)&0xFF) << "." <<
                    ((soap.ip >> 8)&0xFF) << "." << (soap.ip&0xFF) << std::endl;
#endif
                struct soap *tsoap = soap_copy(&soap);
                threads.create_thread(thread_adapter(&process_request, (void*)tsoap));
            }
        }
    }
    catch (std::exception& e)
    {
        std::cerr << "Exception in fannKernel: " << e.what() << std::endl;
    }
    catch (...)
    {
        std::cerr << "Generel exception in fannKernel" << std::endl;
    }
    return 0;
}
